from dbis_tm import Schedule, ConflictGraph, ConflictGraphNode, SyntaxCheck
from tests.scheduletest import ScheduleTest


class Test_TM(ScheduleTest):
    # A tuple denotes: (schedule, is_well_formed, is_serializable)
    unparsed_schedule_tests = [
        ("w1(x)w1(y)r2(u)w2(x)r2(y)r3(x)w2(z)a2r1(z)c1c3", True, True),
        # well-formed, serializable
        ("R1(x)W1(x)r2(x)A1w2(x)C2", True, True),
        # well-formed, serializable
        ("w1(Y)w4(W)w3(Z)w2(X)r1(Z)r1(X)r4(Z)", True, True),
        # well-formed, serializable
        (
            "r1(x)w2(y)r1(x)w3(z)w3(x)r1(y)w1(y)w2(z)w1(z)w3(y)r2(x)c3r2(y)c2w1(y)a1",
            True,
            False,
        ),
        # well-formed, not serializable
        (
            "r1(x)w2(y)r1(x)w3(z)w3(x)r1(y)w1(y)w2(z)w1(z)w3(y)c3r2(y)c2w1(y)c1",
            True,
            False,
        ),
        # well-formed, not serializable
        (
            "r1(x)w2(y)r1(x)w3(z)w3(x)r1(y)w1y)w2(z)w1(z)w3(y)c3r2(y)c2w1(y)c1",
            False,
            False,
        ),
        # malformed
        (
            "r1(x)w2(y)r1(x)w3(z)w3(x)r1(y)w(y)w2(z)w1(z)w3(y)c3r2(y)c2w1(y)c1",
            False,
            False,
        ),
        # malformed
        (
            "r1(x)w2(y)r1(x)w3(z)w3(x)r1(y)w1(yw2(z)w1(z)w3(y)c3r2(y)c2w1(y)c1",
            False,
            False,
        ),
        # malformed
        (
            "r1(x)w2(y)r1(x)w3(z)w3(x)r1(y)1(y)w2(z)w1(z)w3(y)c3r2(y)c2w1(y)c1",
            False,
            False,
        )
        # malformed
    ]

    # A tuple denotes: (schedule, schedule_mod, compare_schedules)
    compare_schedules_test = [
        (
            "w1(x)r2(e)r1(y)",
            "rl2(e)wl1(x)w1(x)r2(e)rl1(y)r1(y)ru1(y)wu1(x)ru2(e)",
            True,
        ),
        (
            "r2(y)w3(x)w1(z)w3(y)r1(x)r2(z)r3(z)c1c2c3",
            "rl2(y)r2(y)wl3(x)w3(x)wl1(z)w1(z)wl3(y)w3(y)rl1(x)r1(x)wu1(z)ru1(x)rl2(z)r2(z)ru2(z)ru2(y)"
            "rl3(z)r3(z)wu3(x)wu3(y)ru3(z)c1c2c3",
            True,
        ),
    ]

    def test_schedule_parsing(self):
        """
        tests parse_schedule(unparsed_schedule)
        """
        for (schedule, is_well_formed, _), i in zip(
            self.unparsed_schedule_tests, range(0, len(self.unparsed_schedule_tests))
        ):
            _, msg = Schedule.parse_schedule(schedule)
            self.assertEqual(is_well_formed, msg == "", f"Schedule {i}:")

    def test_compare_schedules(self):
        """
        tests check_schedule(schedule, schedule_mod)
        """
        for (schedule, schedule_mod, result), _i in zip(
            self.compare_schedules_test, range(len(self.compare_schedules_test))
        ):
            parsed, msg = Schedule.parse_schedule(schedule)
            parsed_mod, msg_mod = Schedule.parse_schedule(schedule_mod)
            returned = Schedule.is_operations_same(parsed, parsed_mod)
            # returned =len(problems)==0
            self.assertEqual(returned, result)

    def testEdgeLessConflictGraph(self):
        """
        test the content of an empty graph
        """
        g_1 = ConflictGraph()
        t1 = ConflictGraphNode(1)
        t2 = ConflictGraphNode(2)
        self.assertTrue(g_1.isEmpty())
        gvMarkup = g_1.get_graphviz_graph()
        debug = False
        if debug:
            print(gvMarkup)
        self.assertTrue(
            """{
	graph [label="Konfliktgraph "]
}"""
            in str(gvMarkup)
        )
        g_1.add_edge(t1, t2)
        gvMarkup = g_1.get_graphviz_graph()
        if debug:
            print(gvMarkup)
        self.assertTrue("t1 -> t2" in str(gvMarkup))

    def testConfSyntaxCheck(self):
        """
        test the SyntaxCheck functionality for Conflicts
        """
        s1_conf = {("w_2(x)", "r_1(x)"), ("w_1(z)", "w_2(z)")}
        s2_conf = {
            ("r_2(x)", "w_3(x)"),
            ("w_1(y)", "r_2(y)"),
            ("w_1(y)", "w_2(y)"),
            ("w_1(y)", "w_3(y)"),
            ("w_3(z)", "w_1(z)"),
            ("w_3(z)", "w_2(z)"),
            ("r_2(y)", "w_3(y)"),
            ("w_2(y)", "w_3(y)"),
            ("w_2(y)", "r_1(y)"),
            ("w_1(z)", "w_2(z)"),
            ("w_3(y)", "r_1(y)"),
            ("w_3(y)", "w_2(y)"),
            ("r_1(y)", "w_2(y)"),
        }
        conf_err1 = []
        conf_err2 = {}
        conf_err3 = "Garbage"
        conf_err4 = {("a"), ("b", "c", "e")}
        conf_err5 = {("a_3(x)", "b")}
        debug = False
        expectedList = [
            None,
            None,
            "[] ist kein Set",
            None,
            "Garbage ist kein Set",
            (
                "Das Tupel ('b', 'c', 'e') von {('b', 'c', 'e'), 'a'} ist kein Paar",
                "Das Tupel a von {'a', ('b', 'c', 'e')} ist kein Paar",
            ),
            "Das Tupel ('a_3(x)', 'b') von {('a_3(x)', 'b')} hat keine korrekte Syntax",
        ]
        for i, conf in enumerate(
            [s1_conf, s2_conf, conf_err1, conf_err2, conf_err3, conf_err4, conf_err5]
        ):
            msg = SyntaxCheck.check_conf_set_syntax(conf)
            if debug:
                print(f"{i}:{msg}", "test")
            expected = expectedList[i]
            if isinstance(expected, tuple):
                expected1, expected2 = expected
                self.assertTrue(expected1 == msg or expected2 == msg)
            else:
                self.assertEqual(expected, msg)

    def testScheduleSyntaxCheck(self):
        """
        test the SyntaxCheck functionality for Schedules
        """
        schedule_1 = "w_2(x) w_1(z) r_2(y) r_1(x) r_3(z) w_3(x) w_1(y) c_1 c_2 c_3"
        schedule_2 = "r_2(z) w_1(y) r_3(z) r_2(y) r_1(x) w_2(y) w_3(x) c_1 c_2 c_3"
        schedule_3 = "r_1(x) w_2(z) w_3(y) w_2(x) r_3(z) r_1(y) r_2(y) c_1 c_2 c_3"
        schedule_4 = "w2(x)"
        schedule_5 = "rl_1(x) r_1(x) wl_2(z) w_2(z)  wl_3(y) w_3(y) wl_2(x) w_2(x) rl_3(z) r_3(z) wu_3(y) ru_3(z) rl_1(y) r_1(y) ru_1(x) ru_1(y) rl_2(y) r_2(y) wu_2(z) wu_2(x) ru_2(y) c_1 c_2 c_3"
        schedule_6 = "wl_2(x) rl_2(y) w_2(x) wu_2(x) r_2(y) ru_2(y) c_2 wl_1(z) rl_1(x) wl_1(y) w_1(z) wu_1(z) r_1(x) ru_1(x) rl_3(z) wl_3(x) r_3(z) ru_3(z) w_3(x) wu_3(x) c_3 w_1(y) wu_1(y) c_1"
        schedule_err2 = ""
        expected = [
            None,
            None,
            None,
            None,
            None,
            None,
            "Leerer Schedule kann keine Lösung sein",
        ]
        debug = False
        for i, schedule in enumerate(
            [
                schedule_1,
                schedule_2,
                schedule_3,
                schedule_4,
                schedule_5,
                schedule_6,
                schedule_err2,
            ]
        ):
            msg = SyntaxCheck.check_schedule_syntax(schedule)
            if debug:
                print(msg)
            self.assertEqual(expected[i], msg)



"""
Created 2022-05

@author: Lara
@author: Marc
@author: Wolfgang
"""
from __future__ import annotations

import itertools
import sys, re
from enum import Enum, EnumMeta
from typing import Union
from graphviz import Digraph


class OperationTypeMeta(EnumMeta):
    """
    meta class for Operation type
    """

    def __contains__(self, item):
        """
        check that the item is contained in my member values
        """
        return item in [v.value for v in self.__members__.values()]


class OperationType(Enum, metaclass=OperationTypeMeta):
    """
    The kind / type of operation for an Operation
    """

    READ = "r"
    READ_LOCK = "rl"
    READ_UNLOCK = "ru"
    WRITE = "w"
    WRITE_LOCK = "wl"
    WRITE_UNLOCK = "wu"


# class Transaction:
# there is no Transaction class yet since we only need the
# transaction number
# feel free to add this later ...


class Operation:
    """
    I am a step of a transaction
    """

    def __init__(
        self, op_type: OperationType, tx_number: int, resource: str, index: int
    ):
        """
        Constructor

        Args:
            op_type(Operation): the kind of operation
            tx_number(int): link to my Transaction
            resource(str): link to my data object the operation will be applied on
            index(int): my position in the schedule
        """
        self.op_type = op_type
        self.tx_number = tx_number
        self.resource = resource
        self.index = index

    def __repr__(self):
        return f"{self.op_type.value}{self.tx_number}({self.resource})"

    def __eq__(self, obj):
        return (
            isinstance(obj, Operation)
            and self.op_type == obj.op_type
            and self.tx_number == obj.tx_number
            and self.resource == obj.resource
        )

    def __sr__(self, obj):
        """True if operations of same trans and on same resource"""
        return (
            isinstance(obj, Operation)
            and self.tx_number == obj.tx_number
            and self.resource == obj.resource
        )

    def __same__(self, obj):
        """Is the same operation"""
        return (
            isinstance(obj, Operation)
            and self.op_type == obj.op_type
            and self.tx_number == obj.tx_number
            and self.resource == obj.resource
            and self.index == obj.index
        )


class Schedule:
    """
    I am a container for
        a list of operations,
        a set of resources,
        a map of aborts and commits,
        and a count of transactions
    """

    def __init__(
        self,
        operations: list[Operation],
        resources: set[str],
        tx_count: int,
        aborts: dict,
        commits: dict,
    ):
        """
        Constructor:

        Args:
            operations(list[Operation]): the kind of operation
            resources(set[str]): link to my Transaction
            tx_count(int): link to my data object the operation will be applied on
            aborts(dict): my position in the schedule
            commits(dict): my position in the schedule
        """
        self.operations = operations
        self.resources = resources
        self.tx_count = tx_count
        self.aborts = aborts
        self.commits = commits

    def __repr__(self):
        return (
            f"Schedule[operations: {self.operations}, resources: {self.resources}, tx_count: {self.tx_count}, "
            f"aborts: {self.aborts}, commits: {self.commits}]"
        )

    def active(self) -> list[int]:
        """Returns the still active transactions."""
        return [
            i
            for i in range(1, self.tx_count + 1)
            if i not in self.aborts.keys() and i not in self.commits.keys()
        ]

    def next_index(self):
        """Returns the next unused index"""
        if self.operations:
            return (
                max(
                    [0]
                    + [self.operations[-1].index]
                    + list(self.aborts.values())
                    + list(self.commits.values())
                )
                + 1
            )
        else:
            return 1

    def op_trans(self, transaction: int) -> int:
        """Returns how many operations one transaction perfomed"""
        return max(
            0,
            len([op for op in self.operations if op.tx_number == transaction]),
        )

    @classmethod
    def sanitize(cls, schedule: str) -> str:
        """
        return a sanitized schedule

        Args:
            schedule(str): the plain input schedule using underscores and whitespaces

        Returns:
            str: the sanitized schedule with underscores and whitespaces removed
        """
        for removeChar in [" ", "_", "\t", "\n"]:
            schedule = schedule.replace(removeChar, "")
        return schedule

    @classmethod
    def parse_schedule(cls, schedule_str: str) -> tuple[Schedule, str]:
        """
        Parse the given string to a schedule.

        Returns:
            Created Schedule object
            In case of error, the unparseable part is returned. Else an empty string is returned
        """

        # Sanitize input
        schedule_str = Schedule.sanitize(schedule_str)

        parsed_schedule = Schedule([], set(), 0, {}, {})
        tx = set()
        index = 0
        i = 0
        while i < len(schedule_str):
            curr_char = schedule_str[i].lower()
            next_char = schedule_str[i + 1].lower()

            if curr_char + next_char in OperationType:
                operation_type = OperationType(curr_char + next_char)
                index += 1
                i += 2
            elif curr_char in OperationType:
                operation_type = OperationType(curr_char)
                index += 1
                i += 1
            elif curr_char == "c":
                index += 1
                parsed_schedule.commits[int(next_char)] = index

                i += 2
                continue
            elif curr_char == "a":
                index += 1
                parsed_schedule.aborts[int(next_char)] = index
                i += 2
                continue
            else:
                p1 = max(i - 2, 0)
                p2 = min(i + 5, len(schedule_str) - 1)
                return parsed_schedule, schedule_str[p1:p2]

            tx_number = schedule_str[i].lower()
            if not tx_number.isdigit():
                p1 = max(i - 2, 0)
                p2 = min(i + 5, len(schedule_str) - 1)
                return parsed_schedule, schedule_str[p1:p2]
            tx.add(tx_number)
            i += 2

            resource = schedule_str[i].lower()
            if not resource.isalpha():
                p1 = max(i - 2, 0)
                p2 = min(i + 5, len(schedule_str) - 1)
                return parsed_schedule, schedule_str[p1:p2]
            parsed_schedule.resources.add(resource)
            i += 2

            parsed_schedule.operations.append(
                Operation(operation_type, int(tx_number), resource, index)
            )

        parsed_schedule.tx_count = len(tx)
        return parsed_schedule, ""

    @classmethod
    def parse_string(cls, schedule: Schedule) -> tuple[str, str]:
        """
        Parse a given schedule into a string.
        Only works if each transaction is concluded.

        Returns:
            - Parsed string of this schedule,
            - And a error message if somethings wrong
        """
        schedule_str = ""
        abort = schedule.aborts
        commit = schedule.commits
        op_len = len(schedule.operations)
        op_counter = 0
        for i in range(1, op_len + schedule.tx_count + 1):
            if op_counter < op_len:
                operation = schedule.operations[op_counter]
            if op_counter < op_len and operation.index == i:
                op_counter += 1
                schedule_str += (
                    operation.op_type.value
                    + str(operation.tx_number)
                    + "("
                    + operation.resource
                    + ") "
                )
            else:
                trc = list(
                    filter(
                        lambda trans: commit.get(trans) == i,
                        range(1, schedule.tx_count + 1),
                    )
                )
                tra = list(
                    filter(
                        lambda trans: abort.get(trans) == i,
                        range(1, schedule.tx_count + 1),
                    )
                )
                if trc:
                    # get transaction if index from dict
                    schedule_str += "c" + str(trc[0]) + " "
                elif tra:
                    # get transaction of index from dict
                    schedule_str += "a" + str(tra[0]) + " "
                else:
                    return schedule_str, "The index: " + str(i) + "is not given."
        return schedule_str, ""

    @classmethod
    def is_operations_same(
        cls, schedule: Union[Schedule, str], mod_schedule: Union[Schedule, str]
    ) -> bool:
        """
        Checks whether the two  given schedules do have the same operations.

        Gets:
            schedule: 'original' schedule (without locks and unlocks)
            schedule_mod: modified schedule (with locks and unlocks)

        Returns:
           bool: True if no problems
        """

        problems = Schedule.check_operations_same(schedule, mod_schedule)
        return len(problems) == 0

    @classmethod
    def check_operations_same(
        cls, schedule: Union[Schedule, str], mod_schedule: Union[Schedule, str]
    ) -> list:
        """
        Checks whether the two  given schedules do have the same operations.

        Gets:
            schedule: 'original' schedule (without locks and unlocks)
            schedule_mod: modified schedule (with locks and unlocks)

        Returns:
           list of problems
        """
        if isinstance(schedule, str):
            schedule = Schedule.parse_schedule(schedule)
            assert not schedule[1]
            schedule = schedule[0]
        if isinstance(mod_schedule, str):
            mod_schedule = Schedule.parse_schedule(mod_schedule)
            assert not mod_schedule[1]
            mod_schedule = mod_schedule[0]
        problems = []
        org_operations = list(
            filter(
                lambda op: op.op_type in [OperationType.READ, OperationType.WRITE],
                mod_schedule.operations,
            )
        )
        for x in org_operations:
            if x in schedule.operations:
                continue
            else:
                problems.append(x)
        for y in schedule.operations:
            if y in org_operations:
                continue
            else:
                problems.append(y)
        for i in range(1, schedule.tx_count + 1):
            trans_op_mod = list(
                filter(
                    lambda op: op.op_type in [OperationType.READ, OperationType.WRITE]
                    and op.tx_number == i,
                    mod_schedule.operations,
                )
            )
            trans_op_org = list(
                filter(lambda op: op.tx_number == i, schedule.operations)
            )
            if not (trans_op_mod == trans_op_org):
                problems.append(f"{trans_op_mod} != {trans_op_org} at {i}")
        return problems


class ConflictGraph:
    """
    a conflict graph
    """

    def __init__(self, labelPostfix=""):
        """
        constructor

        Args:
            labelPostfix(str): the postfix for the label to be used
        """
        self.nodes = set()
        self.edges = set()
        cglabel = f"Konfliktgraph {labelPostfix}"
        self.digraph = Digraph(
            "Konfliktgraph",
            "generiert von DBIS VL UB 8 TM.ConflictGraph",
            graph_attr={"label": cglabel},
        )

    def isEmpty(self):
        return len(self.nodes) == 0

    def __eq__(self, obj):
        return (
            isinstance(obj, ConflictGraph)
            and self.nodes == obj.nodes
            and self.edges == obj.edges
        )

    def get_graphviz_graph(self):
        return self.digraph

    def add_edge(self, t1: ConflictGraphNode, t2: ConflictGraphNode) -> None:
        self.nodes.add(t1)
        self.nodes.add(t2)
        self.edges.add((t1, t2))
        self.digraph.edge(f"t{t1.tx_number}", f"t{t2.tx_number}")


class ConflictGraphNode:
    """ """

    def __init__(self, tx_number: int):
        self.tx_number = tx_number

    def __eq__(self, obj):
        return isinstance(obj, ConflictGraphNode) and self.tx_number == obj.tx_number

    def __hash__(self):
        return hash(self.tx_number)


class SyntaxCheck:
    """
    I am an interface for checking the syntax of inputs.
    You should not construct me because I am a stateless interface that merely provides static functions.

    Functions:
        check_conf_set_syntax (checks syntax of strings in tuple that denotes conflicting operations)
    """

    def __init__(self):
        raise TypeError("Cannot create 'SyntaxCheck' instances.")

    @classmethod
    def check_schedule_syntax(cls, schedule: str) -> str:
        """
        check the syntax of the given schedule

        Args:
            schedule(str): the schedule to check

        Returns:
            msg: None if ok else the problem message
        """
        schedule = Schedule.sanitize(schedule)
        syntax_pattern = "([rw][lu]?[1-3][(][a-z][)]|[c][1-3])?"
        p_count = re.findall(syntax_pattern, schedule).count("")
        msg = None
        if schedule == "":
            msg = "Leerer Schedule kann keine Lösung sein"
        if p_count > 1:
            msg = f"Schedule '{schedule}' hat keine korrekte Syntax"
        return msg

    @classmethod
    def check_conf_set_syntax(cls, conf_set: set[tuple[str, str]]) -> str:
        """
        Check syntax of strings in tuple that denotes conflicting operations.

        Returns:
            None if input is formatted according to pattern
            or an error message in case a tuple is formatted incorrectly
        """
        tuple_pattern = "[rw][1-3][(][a-z][)]|[rw]_[1-3][(][a-z][)]"
        if conf_set == {}:
            pass
        elif not isinstance(conf_set, set):
            return f"{conf_set} ist kein Set"
        for t in conf_set:
            if not len(t) == 2:
                return f"Das Tupel {t} von {conf_set} ist kein Paar"
            for s in sorted(list(t)):
                if not re.match(tuple_pattern, s):
                    return f"Das Tupel {t} von {conf_set} hat keine korrekte Syntax"
        return None

    @classmethod
    def check(cls, index, schedule, result) -> str:
        """
        check the given schedule against the given result
        """
        msg = None
        s_parsed, s_problem = Schedule.parse_schedule(schedule)
        result_parsed, result_problem = Schedule.parse_schedule(result)
        problems = Schedule.check_operations_same(s_parsed, result_parsed)
        if not len(problems) == 0:
            msg = f"schedule_{index} enthält unterschiedliche oder nicht alle Operationen aus s{index}"
        return msg
